/*
 * Copyright (C) 2017-2021 Dominique Dhoutaut, Thierry Arrabal, Eugen Dedu.
 *
 * This file is part of BitSimulator.
 *
 * BitSimulator is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * BitSimulator is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with BitSimulator.  If not, see <https://www.gnu.org/licenses/>.
 */

#ifndef AGENTS_PROBA_FLOODING_DYNAMIC_RING_ROUTING_AGENT_H_
#define AGENTS_PROBA_FLOODING_DYNAMIC_RING_ROUTING_AGENT_H_

#include <random>
#include "../routing-agent.h"


//===========================================================================================================
//
//          ProbaFloodingDynamicRingRoutingAgent  (class)
//
//===========================================================================================================
typedef struct {
  int controlSeqNoBig;
  int controlSeqNoSmall;
} contSeqNo_t;

typedef std::pair<int, int> datapacket_t; //src,seqno

class ProbaFloodingDynamicRingRoutingAgent : public RoutingAgent {
 protected:
  map<int,int> alreadySeenPackets;
  map<int, contSeqNo_t> amIonRingMap;     //transmitterId (hostnode), ctrl1seqno, ctrl2seqno
  map<datapacket_t, distance_t> countAckMap; 
  static mt19937_64 *forwardingRNG;
  static set<int> reachability;
  bool needToSendControl;
  distance_t rangeBig, rangeSmall;
  bool initialisationStarted;
  float forwardingProbability;

 public:
  ProbaFloodingDynamicRingRoutingAgent(Node *_hostNode);
  virtual ~ProbaFloodingDynamicRingRoutingAgent();
  virtual void receivePacketFromNetwork(PacketPtr _packet);
  virtual void receivePacketFromApplication(PacketPtr _packet);
  bool amIOnRing(PacketPtr _packet);
  int amIOnRingCount();
  void forwardDataPackets(PacketPtr _packet);
  static void initializeAgent();
  virtual void processAckTimerEvent(PacketPtr _packet);
    using ProbaFloodingDynamicRingAckEvent = CallMethodArgEvent<ProbaFloodingDynamicRingRoutingAgent, PacketPtr,
    &ProbaFloodingDynamicRingRoutingAgent::processAckTimerEvent,
    EventType::RINGPROBA_ACK>;
};



#endif /* AGENTS_PROBA_FLOODING_DYNAMIC_RING_ROUTING_AGENT_H_ */
